#!/usr/bin/env python3
"""
Reads the log file created by the environmental sensing script, converts this
to a Pandas DataFrame and saves it to disk in a compressed Python Pickle file
for later analysis.
"""

import argparse
import bz2
import contextlib
import gzip
import lzma
import os
import time

import pandas as pd


##############################################################################
# command line option handler
##############################################################################

def check_arguments():
    """
    handle command line options

    --------------------------------------------------------------------------
    args : none
    --------------------------------------------------------------------------
    returns
        parser.parse_args()
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='Reads the log file created by the environmental sensing\
        script, converts this to a Pandas DataFrame and saves it to disk in a\
        compressed Python Pickle file for later analysis. The log file is\
        designed to be human-readable in a terminal window, and because of\
        this parts of the same data acquisition are split over several lines\
        with timestamps that may differ by 10ms or so. Any lines with\
        timestamps that fall within 200ms of each other are combined into a\
        single DataFrame row.')
    parser.add_argument(
        'log', nargs=1, metavar='filename',
        help='Specify the log file to convert',
        type=str)

    return parser.parse_args()


##############################################################################
# utilities
##############################################################################

def open_method(infile):
    """
    Return the appropriate function to open the given file based on its
    extension. This will set the binary mode correctly for compressed files.

    --------------------------------------------------------------------------
    args
        infile : string
            filename of the log file to read
    --------------------------------------------------------------------------
    returns : function
    --------------------------------------------------------------------------
    """
    infile_extension = os.path.splitext(infile)[-1].lower()
    methods = {'.bz2': bz2.open, '.gz': gzip.open,
               '.xz': lzma.open, '.z': gzip.open}

    return methods.get(infile_extension, open)


def get_data_from_log(infile):
    """
    Extracts data from log, yields sensor readings as they are read.

    log entries look like this:

    2020-02-20 12:44:46,927 : INFO : BME280   21.5 °C , METEOMK2-12E3A7   21.2 °C
    2020-02-20 12:44:46,928 : INFO : BME280   31.4 RH%, METEOMK2-12E3A7   31.0 RH%
    2020-02-20 12:44:46,928 : INFO : BME280  999.7 hPa, METEOMK2-12E3A7 1000.0 hPa
    2020-02-20 12:44:52,102 : INFO : BME280   21.5 °C , METEOMK2-12E3A7   21.2 °C
    2020-02-20 12:44:52,108 : INFO : BME280   31.4 RH%, METEOMK2-12E3A7   31.0 RH%
    2020-02-20 12:44:52,108 : INFO : BME280  999.7 hPa, METEOMK2-12E3A7 1000.0 hPa

    --------------------------------------------------------------------------
    args
        infile : string
            filename of the log file to read
    --------------------------------------------------------------------------
    yields
        row_acc : dict
            e.g. {'timestamp': 1603819534.675, 'bme280_0 °C': 19.8,
                  'bme280_1 °C': 19.2, 'bme280_2 °C': 20.6, ...}
    --------------------------------------------------------------------------
    """

    # '2019-10-04 14:17:12,892'
    # note that %f is a value in microseconds, and the figure given in the
    # string is in milliseconds, so later, the string will be padded with
    # three trailing zeroes to correct this
    log_timestamp_format = '%Y-%m-%d %H:%M:%S,%f,%Z'

    # handle opening of compressed log files
    file_read = open_method(infile)

    with file_read(infile) as templog:
        row_acc = dict()
        row = dict()
        for line in templog.readlines():

            # if the log file is compressed, line will be bytes rather than string
            with contextlib.suppress(AttributeError):
                line = line.decode('utf-8')

            # extract timestamp and convert it to seconds since epoch
            #
            # line = '2020-02-26 17:41:56,861 : INFO : BME280   14.2 °C'
            # timestamp, data = ['2020-02-26 17:41:56,861', 'BME280   14.2 °C']
            try:
                timestamp_raw, datatext = [x.strip() for x in line.split(' : INFO : ')]
            except ValueError:
                continue

            fraction = float('.' + timestamp_raw.rsplit(',')[-1])
            timestamp = timestamp_raw + '000,UTC'
            epoch = time.mktime(time.strptime(timestamp, log_timestamp_format)) + fraction

            # extract for pandas

            # convert this line to dict entries
            sensors = [x.strip().split() for x in datatext.split(',')]
            for sensor in sensors:
                if len(sensor) == 3:
                    row[f'{sensor[0]} {sensor[2]}'] = float(sensor[1])

            # Merge all sequential data acquisition log lines with timestamps
            # within 0.5s into the same DataFrame row, but only if no data
            # would be overwritten in the process. An overwrite could
            # plausibly occur if logs are combined from more than one machine
            # where the same sensor type is in use.
            try:
                dissimilar_acquisition_time = abs(epoch - row_acc['timestamp']) > 0.5
            except KeyError:
                # first line of log file
                row_acc['timestamp'] = epoch
            else:
                conflict = set(row).intersection(set(row_acc))

                if dissimilar_acquisition_time or conflict:
                    # start new acquisition period
                    yield row_acc
                    row_acc.clear()
                    row_acc['timestamp'] = epoch
            finally:
                row_acc.update(row)
                row.clear()


##############################################################################
# main
##############################################################################

def main():
    """
    Reads the log file created by the environmental sensing script, converts
    this to a Pandas DataFrame and saves it to disk as a compressed Python
    Pickle file for later analysis.
    """

    args = check_arguments()
    infile = args.log[0]
    print(f'reading {infile}')

    pdat = pd.DataFrame(dict(x) for x in get_data_from_log(infile))

    outfile = f'{os.path.splitext(infile)[0]}_pandas_csv_zip.dat'
    print(f'writing {outfile}')
    pdat.to_csv(outfile, index=False, compression='zip')


##############################################################################
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Thermo Neslab ULT-80 Low Temperature Bath/Circulator RS232.

The controller looks quite primitive and doesn't use line terminators like
other devices in the test setup. Presumably it just looks for the 0xCA start
byte and ignores anything else when it's sent something unpredictable.

If this turns out not to be the case, we can read the list of available FTDI
serial port devices, then exclude any serial ports found in cache.json, then
leave the user to choose which of the remaining serial ports to use.

Refer to the following for RS232 command sequences:
https://www.atecorp.com/atecorp/media/pdfs/data-sheets/neslab-ult-80-95_manual.pdf

Also see 'setup loop' starting on page 10 to enable RS232.

"Enter the Setup Loop from the Operator's Loop by pressing and holding the NO
 key, then press the NEXT ENTER key. Use the YES/NO keys to adjust the
 values. Press the NEXT ENTER key twice to accept the new value."

tunE?
    YES (up triangle)
HEAt
    NEXT (circle)
rtd
    NEXT (circle)
SP
    NEXT (circle)
HIT
    NEXT (circle)
LoT
    NEXT (circle)
r232
    YES (up triangle)
ON
    YES (up triangle) (only if OFF)
    ENTER (circle)
    ENTER (circle)
bAUd
    YES (up) or NO (down) as appropriate - 9600
    ENTER (circle)
    ENTER (circle)
dAtA
    YES (up) or NO (down) as appropriate - 8
    ENTER (circle)
    ENTER (circle)
StoP
    YES (up) or NO (down) as appropriate - 1
    ENTER (circle)
    ENTER (circle)
Par
    YES (up) or NO (down) as appropriate - N
    ENTER (circle)
    ENTER (circle)
Stor (store changes)
    YES (up arrow)

"NOTE: Should you desire to return to the temperature display and abort all
 changes, keep pressing the NEXT ENTER until the display reads Stor, then
 press NO."
"""

import argparse
import sys

import serial
import serial.tools.list_ports as stlp

from mmcb import common


##############################################################################
# command line option handler
##############################################################################

def check_arguments():
    """
    Handle command line options.

    --------------------------------------------------------------------------
    args : none
    --------------------------------------------------------------------------
    returns : class argparse.ArgumentParser
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='Thermo Neslab ULT-80 Low Temperature Bath/Circulator\
        RS232 tool.')

    group1 = parser.add_mutually_exclusive_group()
    group1.add_argument(
        '-i', '--read_internal_temperature',
        action='store_true',
        help='Read Internal Temperature.')
    group1.add_argument(
        '-r', '--read_setpoint',
        action='store_true',
        help='Read Setpoint (control point).')
    group1.add_argument(
        '-s', '--set_setpoint', nargs=1, metavar='degrees_c',
        help='Set Setpoint (control point).',
        default=None)

    return parser.parse_args()


##############################################################################
# detect ult80 serial port and basic operation
##############################################################################

def find_ult80(port, config):
    """
    Return serial port configuration if the serial port was successfully
    opened.

    --------------------------------------------------------------------------
    args
        port : string
            port name
        config : dict
            serial port configuration
    --------------------------------------------------------------------------
    returns
        retval :
            settings : dict (if port could be opened) or None otherwise
                Note that the dictionary does not contain the serial port
                identifier.
    --------------------------------------------------------------------------
    """
    ser = serial.Serial()
    ser.apply_settings(config)
    ser.port = port

    try:
        ser.open()
    except (OSError, serial.SerialException):
        sys.exit(f'could not open port {port}, exiting.')
    else:
        settings = ser.get_settings()
        ser.close()

    return settings


##############################################################################
# data structures
##############################################################################

class Ult80:
    """
    Thermo Neslab ULT-80 Low Temperature Bath/Circulator RS232.
    """
    _lead_char = 0xca
    _address = 0x0001
    _nbits = 16

    def __init__(self, ser):
        self.ser = ser

    ##########################################################################
    # public functions
    #
    # key:
    #
    # qb = qualifier byte
    #     0x10 : 0.1 precision, no units of measure
    #     0x20 : 0.01 precision, no units of measure
    #     0x11 : 0.1 precision, °C units
    #
    # Example: The temperature of 45.6 °C would be represented by the
    # qualifier 0x11, followed by the 2 bytes 0x01C8 (456 decimal).
    #
    # d1,d2 = 16-bit signed integer of the value being sent or received
    #
    # cs = the checksum of the string
    #
    # "All commands must be entered in the exact format shown in the tables on
    #  the following pages... Controller responses are either the requested
    #  data or an error message. The controller response must be received
    #  before sending the next command."
    #
    # "The host sends a command embedded in a single communications packet,
    #  then waits for the controller’s response. If the command is not
    #  understood or the checksums do not agree, the controller responds with
    #  an error command. Otherwise, the controller responds with the
    #  requested data. If the controller fails to respond within 1 second,
    #  the host should re-send the command."
    ##########################################################################

    def read_internal_temperature(self):
        """
        FUNCTION      : Read Internal Temperature
        MASTER SENDS  : CA 00 01 20 00 DE
        BATH RESPONDS : CA 00 01 20 03 qb d1 d2 cs

        ----------------------------------------------------------------------
        args : none
        ----------------------------------------------------------------------
        returns : float (or None if the checksum was incorrect)
        ----------------------------------------------------------------------
        """
        command = 0x2000
        self.ser.write(self._build_command(command))

        return self._read_response()

    def read_setpoint_control_point(self):
        """
        FUNCTION      : Read Setpoint (control point)
        MASTER SENDS  : CA 00 01 70 00 8E
        BATH RESPONDS : CA 00 01 70 03 qb d1 d2 cs

        ----------------------------------------------------------------------
        args : none
        ----------------------------------------------------------------------
        returns : float (or None if the checksum was incorrect)
        ----------------------------------------------------------------------
        """
        command = 0x7000
        self.ser.write(self._build_command(command))

        return self._read_response()

    def set_setpoint_control_point(self, deg_c):
        """
        FUNCTION      : Set Setpoint (control point)
        MASTER SENDS  : CA 00 01 F0 02 d1 d2 cs
        BATH RESPONDS : CA 00 01 F0 03 qb d1 d2 cs

        ----------------------------------------------------------------------
        args
            deg_c : float
        ----------------------------------------------------------------------
        returns
            decoded_value : float (or None if the checksum was incorrect)
        ----------------------------------------------------------------------
        """
        command = 0xf002
        self.ser.write(self._build_command(command, deg_c))

        return self._read_response()

    ##########################################################################
    # support functions
    ##########################################################################

    def _build_command(self, command, deg_c=None):
        """
        Build the sequence of bytes expected by the ULT-80 for the required
        command.

        ----------------------------------------------------------------------
        args
            command : int
                2-byte value
            deg_c : float
                temperature in degrees C
        ----------------------------------------------------------------------
        returns : bytes
            Sequence of bytes expected by the ULT-80 for the required command.
        ----------------------------------------------------------------------
        """
        lead_char = [self._lead_char]
        address = list(self._address.to_bytes(2, 'big'))
        command = list(command.to_bytes(2, 'big'))

        data = address + command
        if deg_c is not None:
            temperature = self._encode_temperature(deg_c)
            data += temperature

        return bytes(lead_char + data + [self._create_checksum(data)])

    def _read_response(self, debug_data=None):
        """
        Generic read response, e.g. CA 00 01 F0 03 11 01 2C CD

        ----------------------------------------------------------------------
        args : none
        ----------------------------------------------------------------------
        returns
            decoded_value : float (or None if the checksum was incorrect)
        ----------------------------------------------------------------------
        """
        # read the first five bytes, so the last byte contains the number of
        # following data bytes - 3 in this instance - then the remaining
        # checksum byte.
        # e.g. CA 00 01 F0 03
        if debug_data is None:
            response_part_1 = self.ser.read(size=5)
        else:
            response_part_1 = debug_data[:5]

        if not response_part_1:
            return None

        if debug_data is None:
            bytes_to_read = response_part_1[-1] + 1
            response_part_2 = self.ser.read(size=bytes_to_read)
        else:
            response_part_2 = debug_data[5:]

        # verify response
        response = response_part_1 + response_part_2

        if self._verify_checksum(response):
            qualifier_byte = response_part_2[0]
            encoded_value = response_part_2[1:3]
            decoded_value = self._decode_value(encoded_value, qualifier_byte=qualifier_byte)
        else:
            decoded_value = None

        return decoded_value

    def _twos_complement(self, value):
        """
        Generates the two's complement of the given integer, handles negative
        numbers. Reversible.

        ----------------------------------------------------------------------
        args
            value : int
        ----------------------------------------------------------------------
        returns
            value : int
        ----------------------------------------------------------------------
        """
        if value < 0:
            value = (1 << self._nbits) + value
        else:
            if (value & (1 << (self._nbits - 1))) != 0:
                # most-significant bit set (sign)
                value = value - (1 << self._nbits)

        return value

    def _encode_temperature(self, temperature):
        """
        Express a temperature value (degrees C) as a 16-bit signed value then
        convert to a pair of bytes in high/low order.

        ----------------------------------------------------------------------
        args
            temperature : float
                Temperature in degrees celsius, positive or negative, only one
                decimal place supported.
        ----------------------------------------------------------------------
        returns : list of two ints
            Each value is a byte value in high/low order.
                e.g.
                    encode_temperature(30) returns
                    [1, 44] or ['0x1', '0x2c']

                    encode_temperature(-10.5) returns
                    [255, 151] or ['0xff', '0x97']
        ----------------------------------------------------------------------
        """
        value = self._twos_complement(int(temperature * 10))

        return list(value.to_bytes(2, 'big'))

    def _decode_value(self, value, qualifier_byte=0x11):
        """
        Decode two-byte value received from ULT-80 into human-readable value.

        ----------------------------------------------------------------------
        args
            value : list of two ints
                contains individual byte values
            qualifier_byte : int
                This will be one of three values:
                    0x10 : 0.1 precision, no units of measure
                    0x20 : 0.01 precision, no units of measure
                    0x11 : 0.1 precision, °C units
        ----------------------------------------------------------------------
        returns
            val : float
        ----------------------------------------------------------------------
        """
        if qualifier_byte not in {0x10, 0x11, 0x20}:
            return None

        denominator = 10 if qualifier_byte in {0x10, 0x11} else 100
        val = self._twos_complement(int.from_bytes(value, 'big')) / denominator

        return val

    @staticmethod
    def _create_checksum(data):
        """
        Generate single byte checksum from data.

        Bitwise inversion of the 1 byte sum of bytes beginning with the most
        significant address byte and ending with the byte preceding the
        checksum. (To perform a bitwise inversion, "exclusive OR" the one
        byte sum with FF hex.)

        ----------------------------------------------------------------------
        args
            data : list of ints
                contains individual byte values, list should only contain data
                for checksumming - no other byte values
        ----------------------------------------------------------------------
        returns : int
            a single byte value
        ----------------------------------------------------------------------
        """
        return (sum(data) & 0xff) ^ 0xff

    def _verify_checksum(self, entire_response):
        """
        Verify checksum from received data.

        ----------------------------------------------------------------------
        args
            entire_response : list of ints
                Contains individual byte values, list should contain all
                received data.
        ----------------------------------------------------------------------
        returns : boolean
            True if checksums match, False otherwise.
        ----------------------------------------------------------------------
        """
        _lead_char, *data, checksum = entire_response

        return self._create_checksum(data) == checksum


##############################################################################
# main
##############################################################################

def main():
    """
    Thermo Neslab ULT-80 Low Temperature Bath/Circulator RS232.
    """
    args = check_arguments()

    # obtain available ports
    all_ports = {com.device for com in stlp.comports() if common.rs232_port_is_valid(com)}
    if not all_ports:
        sys.exit('No usable serial ports found')

    # find ports already allocated to other devices
    cache = common.cache_read()

    # If a port has None in its settings field, then it was not identified by
    # detect.py, and can therefore still be considered for use by this
    # script.
    identified_ports_in_cache = {k for k, v in cache.items() if v[0] is not None}

    # usable ports for this script
    available_ports = all_ports - identified_ports_in_cache

    if len(available_ports) > 1:
        sys.exit(f'Too many ports to choose from: {available_ports}')
    elif not available_ports:
        sys.exit('No usable serial ports available')

    port = available_ports.pop()
    print(f'using port {port}')
    config = {'baudrate': 9600, 'bytesize': 8, 'parity': 'N', 'stopbits': 1,
              'xonxoff': False, 'dsrdtr': False, 'rtscts': False, 'timeout': 1,
              'write_timeout': 1, 'inter_byte_timeout': None}

    settings = find_ult80(port, config)
    if settings is None:
        sys.exit('Serial port could not be opened')

    with serial.Serial(port=port) as ser:
        ser.apply_settings(settings)

        ult80 = Ult80(ser)

        if args.read_internal_temperature:
            rval = ult80.read_internal_temperature()
            if rval is None:
                print('no response from device')
            else:
                print(f'read_internal_temperature, returned value {rval}°C')

        if args.read_setpoint:
            rval = ult80.read_setpoint_control_point()
            if rval is None:
                print('no response from device')
            else:
                print(f'read_setpoint_control_point, returned value {rval}°C')

        if args.set_setpoint:
            rval = ult80.set_setpoint_control_point(float(args.set_setpoint[0]))
            if rval is None:
                print('no response from device')
            else:
                print(f'set_setpoint_control_point, returned value {rval}°C')


##############################################################################
if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="mmcb-avt",
    version="0.0.61",
    author="Alan Taylor, Manex Ormazabal",
    author_email="avt@hep.ph.liv.ac.uk",
    maintainer="Alan Taylor",
    maintainer_email="avt@hep.ph.liv.ac.uk",
    description="ATLAS ITK Pixels Multi-Module Cycling Box environmental monitoring/control.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.ph.liv.ac.uk/avt/atlas-itk-pmmcb",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=[
        "numpy",
        "pandas",
        "matplotlib",
        "zmq",
        "pyserial>=3.5",
        "smbus",
        "tables",
        "Adafruit-Blinka",
        "adafruit-circuitpython-ads1x15",
        "adafruit-circuitpython-bme680",
        "adafruit-circuitpython-pcf8591",
        "adafruit-circuitpython-shtc3",
        "sparkfun-qwiic-tca9548a",
        "yoctopuce",
    ],
    entry_points={
        "console_scripts": [
            "dat2plot = mmcb.dat2plot:main",
            "dat2root = mmcb.dat2root:main",
            "detect = mmcb.detect:main",
            "iv = mmcb.iv:main",
            "liveplot = mmcb.liveplot:main",
            "log2dat = mmcb.log2dat:main",
            "peltier = mmcb.peltier:main",
            "psuset = mmcb.psuset:main",
            "psustat = mmcb.psustat:main",
            "sense = mmcb.sense:main",
            "ult80 = mmcb.ult80:main",
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: POSIX :: Linux",
        "Natural Language :: English",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
    ],
)



#!/usr/bin/env python3
"""
Plots live data outputted by iv.py in a window that persists over multiple
runs of iv.py.

Uses ZeroMQ REQ/REP for communication.
"""

import argparse
import gc
import itertools
import multiprocessing as mp    # Process, Queue
import sys
import threading
import time

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.ticker import EngFormatter
import zmq


##############################################################################
# command line option handler
##############################################################################

def check_arguments():
    """
    handle command line options

    --------------------------------------------------------------------------
    args : none
    --------------------------------------------------------------------------
    returns : bool
        True if the user has used the --forwardbias command line option,
        False otherwise.
    --------------------------------------------------------------------------
    """
    parser = argparse.ArgumentParser(
        description='This is an optional companion script for iv.py. It plots\
        IV data broadcast by iv.py as it is acquired; the plot\
        window persists over multiple runs of iv.py, until it is terminated\
        by the user. This script communicates with iv.py using ZeroMQ on\
        local TCP port 5555.')
    parser.add_argument(
        '--forwardbias',
        action='store_true',
        help='Use this option if iv.py has been called with --forwardbias to\
        allow it to use positive voltages. This will prevent this script from\
        inverting the plot axes.')

    args = parser.parse_args()

    return args.forwardbias


##############################################################################
# threads
##############################################################################

def liveplot(pipeline):
    """
    Receive data packets from iv.py via ZeroMQ REQ/REP, and place them on a
    queue for processing later.

    To avoid dropped data packets ZeroMQ REQ/REP is the pattern used, though
    the desired behaviour is really PUB/SUB. PUB/SUB is unreliable in its
    basic form, and can't be used. With REQ/REP this thread is obliged to
    reply to the message iv.py sends, but since that reply will be discarded
    by iv.py, only send the most minimal reply.

    This activity is run in thread to decouple receiving data packets from the
    live-plotting process.

    --------------------------------------------------------------------------
    args
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
    --------------------------------------------------------------------------
    returns : no explicit return
        pipeline.packet : multiprocessing.Queue()
            shared repository for received data packets
    --------------------------------------------------------------------------
    """
    # configure communication with iv.py
    # Use REP/REQ (with handshaking) instead of PUB/SUB to avoid packet loss
    context = zmq.Context(1)
    socket = context.socket(zmq.REP)
    try:
        socket.bind("tcp://*:5555")
    except zmq.error.ZMQError:
        pass
    else:
        while pipeline.terminate_liveplot.empty():
            # receive message from iv.py
            message = socket.recv_json()

            pipeline.packet.put(message)

            # send a minimal message back to iv.py for handshake
            socket.send_json(None)


##############################################################################
# utilities
##############################################################################

def collect_data_packets(pipeline):
    """
    Given that data packets should arrive at a fairly low rate (one
    data packet per power supply channel per second), in typical usage this
    function is unlikely to return more than one data packet per plot
    interval.

    In larger test environments with several devices under test (and given
    that power supply channels are tested asynchronously), or when using test
    scripts, it's plausible that more than one data packet may arrive per
    plot interval, and this function manages this eventuality.

    Regarding iv.py sending reset packets to clear the plot: this function is
    called by animate() at the interval specified by the call to
    animation.FuncAnimation() (specified with interval=..., the default is
    every 200ms). Therefore any data packets received before a reset will be
    discarded, since there's little point live-plotting data that will
    be visible for 200ms. This is a corner case, since for this to occur
    one invocation of iv.py would have to terminate, and another one start
    within the narrow call interval, which can only be achieved with a test
    script, and not through normal operation.

    --------------------------------------------------------------------------
    args
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
    --------------------------------------------------------------------------
    returns
        packets : iterator
            e.g. ('reset', 'debug 1234567,0.0,-0.0,False')
    --------------------------------------------------------------------------
    """
    packets = []

    # get all packets from the queue
    # this function is the only one that reads from this queue so get() should
    # never block
    while not pipeline.packet.empty():
        packets.append(pipeline.packet.get())

    # discard any packets received before the last reset packet, always
    # retaining the last 'reset' packet itself
    try:
        rindex = -packets[::-1].index('reset') - 1
    except ValueError:
        pass
    else:
        packets = packets[rindex:]

    return packets


##############################################################################
# plot
##############################################################################

def animate(_, pipeline, plotdata, collut, colors, axis, forwardbias):
    """
    Obtain new data from iv.py via ZeroMQ and plot graph with updated data.

    --------------------------------------------------------------------------
    args
        _ : int
            frame counter (ignore)
        pipeline : instance of class Production
            contains all the queues through which the production pipeline
            processes communicate
        plotdata : dict
        collut : dict
        colors : list of tuples (R, G, B, A)
            e.g.
            [(0.12156862745098039, 0.4666666666666667, 0.7058823529411765, 1.0),
             (1.0, 0.4980392156862745, 0.054901960784313725, 1.0), ...]
        axis : matplotlib.axes._subplots.AxesSubplot
        forwardbias : bool
            True if the user has used the --forwardbias command line option,
            False otherwise.
    --------------------------------------------------------------------------
    returns : none
    --------------------------------------------------------------------------
    """
    messages = collect_data_packets(pipeline)
    if not messages:
        return

    # process data packets received this interval
    for message in messages:
        try:
            # string, float, float, bool
            ident, volt, curr, rev = message
        except (ValueError, AttributeError):
            if message == 'reset':
                print('packet: start new plot')
                plotdata.clear()
                collut.clear()
        else:
            print(f'packet: {message}')

            suffix = 'return' if rev else 'outbound'
            label = f'{ident} {suffix}'

            # first data point for this power supply: initialise data structure
            if plotdata.get(label) is None:
                plotdata[label] = [[], [], ident]
            if collut.get(ident) is None:
                collut[ident] = None

            plotdata[label][0].append(volt)
            plotdata[label][1].append(curr)

    # plot
    if plotdata:
        # pair line colour with psu channel ident
        for lut, col in zip(collut, itertools.cycle(colors)):
            if collut[lut] is None:
                collut[lut] = col

        axis.clear()
        for label, (volt, curr, pid) in plotdata.items():
            linetype = '.-' if 'outbound' in label else '.--'
            axis.plot(volt, curr, linetype,
                      label=label, linewidth=0.5,
                      markersize=1, color=collut[pid])

        if not forwardbias:
            axis.invert_yaxis()
            axis.invert_xaxis()

        axis.set_xlabel('bias voltage (V)')
        axis.set_ylabel('leakage current (A)')
        axis.set_title('IV Liveplot')
        axis.legend(loc='lower right')
        axis.yaxis.set_major_formatter(EngFormatter(places=1))
        axis.xaxis.set_major_formatter(EngFormatter(places=1))
        axis.yaxis.tick_left()
        axis.xaxis.tick_bottom()
        plt.tight_layout()

    # This is to address Raspberry Pi OS's use of legacy matplotlib 3.0.2
    # which seems to accumulate lots of memory over time. No harm in calling
    # this for other configurations.
    gc.collect()


##############################################################################
# main
##############################################################################

def main():
    """
    Plots live data outputted by iv.py in a window that persists over multiple
    runs of iv.py.
    """
    forwardbias = check_arguments()

    ##########################################################################
    # matplotlib
    ##########################################################################

    # set matplotlib defaults
    matplotlib.rcParams.update({
        # remove chartjunk
        'axes.spines.top': False,
        'axes.spines.right': False,
        # fontsize of the x any y labels
        'axes.labelsize': 'medium',
        # fontsize of the axes title
        'axes.titlesize': 'medium',
        # fontsize of the tick labels
        'xtick.labelsize': 'small',
        'ytick.labelsize': 'small',
        # fontsize of plot-line labels
        'legend.fontsize': 'x-small'})

    fig = plt.figure(num='liveplot.py (receives data from iv.py)')
    axis = fig.add_subplot(1, 1, 1)

    ##############################################################################
    # set up resources for threads
    ##############################################################################

    # define data structures for processes to use
    class Production:
        """Queues to support production pipeline"""
        terminate_liveplot = mp.Queue()
        packet = mp.Queue()

    pipeline = Production()

    # livp: collect sampled data points sent from iv.py
    livp = threading.Thread(target=liveplot, args=(pipeline, ))
    livp.start()

    # allow some time before checking if the thread has exited, to give enough
    # time for binding to the desired port to fail
    time.sleep(0.1)
    if not livp.is_alive():
        sys.exit('liveplot.py may already be running')

    ##############################################################################
    # liveplot
    ##############################################################################

    # create list of colours from a ten-entry colour map
    colors = [plt.cm.tab10(x) for x in range(10)]
    plotdata = dict()
    collut = dict()

    print('waiting for data packets')
    ani = animation.FuncAnimation(fig, animate,
                                  fargs=(pipeline, plotdata, collut, colors, axis, forwardbias))

    plt.show()

    ##########################################################################
    # release resources for threads
    ##########################################################################

    pipeline.terminate_liveplot.put(True)
    livp.join()


##############################################################################
if __name__ == '__main__':
    main()
